import numpy as np

from circleguard import ReplayPath, Mod, Circleguard, order, ReplayMap
from tests.utils import CGTestCase, DELTA, UR_DELTA, RES, FRAMETIME_LIMIT


class TestSnaps(CGTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.r1 = ReplayPath(RES / "corrected_replay1.osr")
        # TODO replays r2 and r5 might not have been properly named, r2 has 0
        # snaps
        cls.r2 = ReplayPath(RES / "legit" / "legit_snaps-1.osr")
        cls.r3 = ReplayPath(RES / "legit" / "legit_snaps-2.osr")
        cls.r4 = ReplayPath(RES / "legit" / "legit_snaps-3.osr")
        cls.r5 = ReplayPath(RES / "legit" / "legit_snaps-4.osr")

    def test_snaps(self):
        snaps = self.cg.snaps(self.r1)

        self.assertEqual(len(snaps), 14)
        # beginning
        self.assertEqual(snaps[0].time, 5103)
        self.assertAlmostEqual(snaps[0].angle, 7.38491, delta=DELTA)
        self.assertAlmostEqual(snaps[0].distance, 16.69009, delta=DELTA)
        # middle
        self.assertEqual(snaps[8].time, 71652)
        self.assertAlmostEqual(snaps[8].angle, 6.34890, delta=DELTA)
        self.assertAlmostEqual(snaps[8].distance, 27.59918, delta=DELTA)
        # end
        self.assertEqual(snaps[13].time, 76502)
        self.assertAlmostEqual(snaps[13].angle, 3.04130, delta=DELTA)
        self.assertAlmostEqual(snaps[13].distance, 21.76919, delta=DELTA)

    def test_snaps_only_on_hitobjs(self):
        r = ReplayMap(221777, 39828)
        snaps = self.cg.snaps(r, only_on_hitobjs=False)
        self.assertEqual(len(snaps), 6)

        # beginning
        self.assertEqual(snaps[0].time, 3410)
        self.assertAlmostEqual(snaps[0].angle, 0.19259, delta=DELTA)
        self.assertAlmostEqual(snaps[0].distance, 44.61642, delta=DELTA)
        # middle
        self.assertEqual(snaps[2].time, 19622)
        self.assertAlmostEqual(snaps[2].angle, 1.87673, delta=DELTA)
        self.assertAlmostEqual(snaps[2].distance, 76.04480, delta=DELTA)
        # end
        self.assertEqual(snaps[5].time, 68833)
        self.assertAlmostEqual(snaps[5].angle, 4.39870, delta=DELTA)
        self.assertAlmostEqual(snaps[5].distance, 8.14900, delta=DELTA)

        snaps = self.cg.snaps(r, only_on_hitobjs=True)
        self.assertEqual(len(snaps), 2)

        self.assertEqual(snaps[0].time, 68822)
        self.assertAlmostEqual(snaps[0].angle, 3.92694, delta=DELTA)
        self.assertAlmostEqual(snaps[0].distance, 8.14900, delta=DELTA)

        self.assertEqual(snaps[1].time, 68833)
        self.assertAlmostEqual(snaps[1].angle, 4.39870, delta=DELTA)
        self.assertAlmostEqual(snaps[1].distance, 8.14900, delta=DELTA)

    def test_legit_snaps(self):
        # these replays have snaps that are picked up when we don't filter for
        # those on hitobjects. Some of them also have snaps that get through
        # that filter anyway. Test mostly exists to alert us to accidental
        # changes in behavior (either good or bad changes).

        snaps = self.cg.snaps(self.r3, only_on_hitobjs=False)
        filtered_snaps = self.cg.snaps(self.r3)
        self.assertEqual(len(snaps), 1)
        self.assertEqual(len(filtered_snaps), 0)

        snaps = self.cg.snaps(self.r4, only_on_hitobjs=False)
        filtered_snaps = self.cg.snaps(self.r4)
        self.assertEqual(len(snaps), 3)
        self.assertEqual(len(filtered_snaps), 1)

    def test_snaps_only_on_hitobjs_accounts_for_time(self):
        # checking that a frame that's in the radius of the nearest hitobj,
        # but isn't within the hitobj's hitwindow and so can't hit the note,
        # is not counted as a snap.
        # This replay previously had 1 snap in it. See
        # https://github.com/circleguard/circleguard/issues/123
        r = ReplayMap(2769844, 448316, mods=Mod.HDHR)
        snaps = self.cg.snaps(r)
        self.assertEqual(len(snaps), 0)

class TestSimilarity(CGTestCase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.stolen1 = ReplayPath(RES / "stolen_replay1.osr")
        cls.stolen2 = ReplayPath(RES / "stolen_replay2.osr")
        cls.legit1 = ReplayPath(RES / "legit" / "legit-1.osr")
        cls.legit2 = ReplayPath(RES / "legit" / "legit-2.osr")
        cls.time_shifted1 = ReplayPath(RES / "stealing" / "stolen-time-shifted-1-1.osr")
        cls.time_shifted2 = ReplayPath(RES / "stealing" / "stolen-time-shifted-1-2.osr")

    @staticmethod
    def add_noise_and_positional_translation_to_replay(replay, pixel_offset, std_deviation):
        mean = [pixel_offset, pixel_offset]
        covariance_matrix = [[std_deviation, std_deviation],[std_deviation,std_deviation]]
        replay.xy = replay.xy + np.random.default_rng().multivariate_normal(mean, covariance_matrix, len(replay.xy))

    def test_cheated(self):
        # taken from http://redd.it/bvfv8j, remodded replay by same user (CielXDLP) from HDHR to FLHDHR
        sim = self.cg.similarity(self.stolen1, self.stolen2)
        self.assertLess(sim, Circleguard.SIM_LIMIT, "Cheated replays were not detected as cheated")

        r1 = self.stolen1
        r2 = self.stolen2
        (earlier, later) = order(r1, r2)

        self.assertAlmostEqual(sim, 2.20867, delta=DELTA, msg="Similarity is not correct")
        self.assertEqual(r1.map_id, r2.map_id, "Replay map ids did not match")
        self.assertEqual(r1.map_id, 1988753, "Replay map id was not correct")
        self.assertEqual(earlier.mods, Mod.HD + Mod.HR, "Earlier replay mods was not correct")
        self.assertEqual(later.mods, Mod.FL + Mod.HD + Mod.HR, "Later replay mods was not correct")
        self.assertEqual(earlier.replay_id, 2801164636, "Earlier replay id was not correct")
        self.assertEqual(later.replay_id, 2805164683, "Later replay id was not correct")
        self.assertEqual(r1.username, r2.username, "Replay usernames did not match")

    def test_cheated_time_shift(self):
        sim = self.cg.similarity(self.time_shifted1, self.time_shifted2, method="similarity")
        self.assertAlmostEqual(sim, 17.30254, delta=DELTA, msg="Similarity is not correct")

        # `similarity` is currently *not* able to detect time shifts. If this
        # changes we want to know! :P
        self.assertGreater(sim, Circleguard.SIM_LIMIT)

        # `correlation` should be able to, though.
        corr = self.cg.similarity(self.time_shifted1, self.time_shifted2, method="correlation")
        self.assertGreater(corr, Circleguard.CORR_LIMIT, "Cheated replays were not detected as cheated")
        self.assertAlmostEqual(corr, 0.99734, delta=DELTA, msg="Correlation is not correct")

    def test_legitimate(self):
        sim = self.cg.similarity(self.legit1, self.legit2)
        self.assertGreater(sim, Circleguard.SIM_LIMIT, "Legitimate replays were detected as stolen")

        r1 = self.legit1
        r2 = self.legit2
        (earlier, later) = order(r1, r2)

        self.assertAlmostEqual(sim, 23.13951, delta=DELTA, msg="Similarity is not correct")
        self.assertEqual(r1.map_id, r2.map_id, "Replay map ids did not match")
        self.assertEqual(r1.map_id, 722238, "Replay map id was not correct")
        self.assertEqual(earlier.mods, Mod.HD + Mod.NC, "Earlier replay mods was not correct")
        self.assertEqual(later.mods, Mod.HD + Mod.DT, "Later replay mods was not correct")
        self.assertEqual(earlier.replay_id, 2157431869, "Earlier replay id was not correct")
        self.assertEqual(later.replay_id, 2309618113, "Later replay id was not correct")
        self.assertEqual(earlier.username, "Crissinop", "Earlier username was not correct")
        self.assertEqual(later.username, "TemaZpro", "Later username was not correct")

    def test_robustness_to_translation(self):
        # copy replay to avoid any missahaps when we mutate the data
        stolen2 = ReplayPath(self.stolen2.path)
        self.cg.load(stolen2)
        TestSimilarity.add_noise_and_positional_translation_to_replay(stolen2, 10, 3)
        sim = self.cg.similarity(self.stolen1, stolen2, method="similarity")
        corr = self.cg.similarity(self.stolen1, stolen2, method="correlation")

        self.assertLess(sim, Circleguard.SIM_LIMIT, "Cheated replays were not detected as cheated with sim")
        self.assertGreater(corr, Circleguard.CORR_LIMIT, "Cheated replays were not detected as cheated with correlation")


class TestUR(CGTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        # TODO check `legit-12.osr` as well once ur calc notelock issues are
        # dealt with, which causes us to be ~3 ur off from its actual ur
        cls.replays = [ReplayPath(RES / "legit" / f"legit-{i}.osr") for i in range(1, 12)]
        cls.urs =          [66.74, 66.56, 242.73, 115.54, 254.56, 90.88, 121.62, 163.01, 207.31, 198.79, 138.25]
        cls.urs_adjusted = [66.73, 63.03, 170.23, 110.76, 182.22, 88.63, 92.53,  104.43, 132.06, 160.68, 104.58]

    def test_ur(self):
        for i, replay in enumerate(self.replays):
            self.assertAlmostEqual(self.cg.ur(replay), self.urs[i],
                delta=UR_DELTA, msg=f"ur differs for replay {i} ({replay})")
            self.assertAlmostEqual(self.cg.ur(replay, adjusted=True),
                self.urs_adjusted[i], delta=UR_DELTA,
                msg=f"adjusted ur differs for replay {i} ({replay})")


class TestFrametime(CGTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.timewarped1 = ReplayPath(RES / "timewarped" / "timewarped-1.osr")
        cls.timewarped2 = ReplayPath(RES / "timewarped" / "timewarped-2.osr")
        cls.timewarped3 = ReplayPath(RES / "timewarped" / "timewarped-3.osr")
        cls.timewarped4 = ReplayPath(RES / "timewarped" / "timewarped-4.osr")

        cls.legit = ReplayPath(RES / "legit" / "legit-1.osr")

    def test_cheated(self):
        replays = [self.timewarped1, self.timewarped2, self.timewarped3, self.timewarped4]
        frametimes = [11.33333, 10.66666, 8, 8.66666]
        for i, replay in enumerate(replays):
            frametime = self.cg.frametime(replay)
            self.assertAlmostEqual(frametime, frametimes[i], delta=DELTA, msg=f"Frametime was wrong for replay {replay}")
            self.assertLess(frametime, FRAMETIME_LIMIT, f"Timewarped replay {replay} was not detected as cheated")

class TestHits(CGTestCase):
    @classmethod
    def setUpClass(cls):
        # don't use cache for this test, it changes xy values slightly
        super().setUpClass(use_cache=False)
        cls.replay1 = ReplayMap(221777, 2757689)

    def test_hits(self):
        hits = self.cg.hits(self.replay1)
        self.assertEqual(len(hits), 1447)

        # beginning
        self.assertEqual(hits[0].t, 22109)
        self.assertSequenceEqual(hits[0].xy.tolist(), [24, 272.8889])
        self.assertEqual(hits[0].hitobject.t, 22110)
        self.assertSequenceEqual(hits[0].hitobject.xy.tolist(), [22, 268])
        self.assertEqual(hits[0].hitobject.radius, 31.104)

        # middle
        self.assertEqual(hits[524].t, 171421)
        self.assertSequenceEqual(hits[524].xy.tolist(), [404, 233.7778])
        self.assertEqual(hits[524].hitobject.t, 171421)
        self.assertSequenceEqual(hits[524].hitobject.xy.tolist(), [409, 227])
        self.assertEqual(hits[524].hitobject.radius, 31.104)

        self.assertEqual(hits[1287].t, 336601)
        self.assertSequenceEqual(hits[1287].xy.tolist(), [233.3333, 241.3333])
        self.assertEqual(hits[1287].hitobject.t, 336593)
        self.assertSequenceEqual(hits[1287].hitobject.xy.tolist(), [243, 259])
        self.assertEqual(hits[1287].hitobject.radius, 31.104)

        # end
        self.assertEqual(hits[1446].t, 385908)
        self.assertSequenceEqual(hits[1446].xy.tolist(), [342.6667, 261.3333])
        self.assertEqual(hits[1446].hitobject.t, 385904)
        self.assertSequenceEqual(hits[1446].hitobject.xy.tolist(), [340, 266])
        self.assertEqual(hits[1446].hitobject.radius, 31.104)



import numpy as np
try:
    from matplotlib import pyplot
except ImportError:
    raise ImportError("matplotlib must be installed to create frametime graphs")

from circleguard.mod import Mod
from circleguard.circleguard import KeylessCircleguard

# classes that import optional modules have to be in their own file, or else
# they get imported and evaluated (and potentially error) when circlecore is
# imported. We can fix this by wrapping the entire definition in a
# try/catch/pass, but then we can't give a nice error message when people try to
# use this class without installing matplotlib. Another file isn't the end of
# the world.

class FrametimeGraph:
    # for any frametimes larger than this, chuck them into a single bin.
    # matplotlib can't really handle that many bins otherwise
    MAX_FRAMETIME = 50

    def __init__(self, replay, cv, figure, show_expected_frametime):
        self.cv = cv
        # figsize is in inches for whatever reason lol
        self.figure = figure or pyplot.figure(figsize=(5, 5))
        self.show_expected_frametime = show_expected_frametime
        conversion_factor_raw = self._conversion_factor(replay)
        self.conversion_factor = conversion_factor_raw if self.cv else 1
        self.expected_frametime = 16 + 2 / 3
        if not self.cv:
            self.expected_frametime /= conversion_factor_raw

        # replay is guaranteed to be loaded when we get it
        cg = KeylessCircleguard()
        # we convert the frametimes manually if necessary later on instead of
        # relying on our conversion in this method, since there are a few
        # oddities about the frametime graph which make it easier this way.
        frametimes = cg.frametimes(replay, cv=False)

        self.figure.suptitle(f"Frametimes for {replay.username} "
            f"+{replay.mods.short_name()} on b/{replay.map_id}")

        self.max_frametime = max(frametimes)
        if self.max_frametime > self.MAX_FRAMETIME:
            self.plot_with_break(frametimes)
        else:
            self.plot_normal(frametimes)

    def plot_normal(self, frametimes):
        frametimes = self.conversion_factor * frametimes
        ax = self.figure.subplots()

        bins = np.arange(0, (self.conversion_factor * self.max_frametime) + 1,
            self.conversion_factor)
        ax.hist(frametimes, bins)
        ax.set_xlabel("Frametime")
        ax.set_ylabel("Count")

        if self.show_expected_frametime:
            ax.axvline(x=self.expected_frametime, color="red")

    # adapted from
    # https://matplotlib.org/examples/pylab_examples/broken_axis.html
    def plot_with_break(self, frametimes):
        # gridspec_kw to make outlier plot smaller than the main one.
        # https://stackoverflow.com/a/35881382
        ax1, ax2 = self.figure.subplots(1, 2, sharey=True,
            gridspec_kw={"width_ratios": [3, 1]})
        ax1.spines["right"].set_visible(False)
        ax2.spines["left"].set_visible(False)
        ax1.set_xlabel("Frametime")
        ax1.set_ylabel("Count")

        ax2.tick_params(left=False)

        low_frametime_truth_arr = frametimes <= self.MAX_FRAMETIME
        low_frametimes = frametimes[low_frametime_truth_arr]
        high_frametimes = frametimes[~low_frametime_truth_arr]

        low_frametimes = self.conversion_factor * low_frametimes
        high_frametimes = self.conversion_factor * high_frametimes

        bins = np.arange(0, (self.conversion_factor * self.MAX_FRAMETIME) + 1,
            self.conversion_factor)
        ax1.hist(low_frametimes, bins)
        # -1 in case high_frametimes has only one frame
        bins = [min(high_frametimes) - 1, self.conversion_factor *
            self.max_frametime]
        ax2.hist(high_frametimes, bins)

        if self.show_expected_frametime:
            ax1.axvline(x=self.expected_frametime, color="red")

    def _conversion_factor(self, replay):
        if Mod.DT in replay.mods:
            return 1 / 1.5
        if Mod.HT in replay.mods:
            return 1 / 0.75
        return 1



# this file is left as an alias so that we can pretend the Mod class comes from
# circlecore, even though we're really using ossapi's Mod class. The Mod class
# used to live in circlecore so consumers may have imported it directly, so
# this linkage needs to exist at least until core 6.0.0. After that we can
# reassess whether it makes sense to leave this or convert our mod imports to
# import from ossapi instead. I think we will want to convert them, but a
# problem for future me.
from ossapi.mod import ModCombination, Mod # pylint: disable=unused-import



from logging import Formatter
from copy import copy
from enum import Enum, IntFlag
from itertools import product, chain, combinations

import numpy as np

from circleguard.mod import Mod


class RatelimitWeight(Enum):
    """
    How much it 'costs' to load a replay from the api.

    :data:`~.RatelimitWeight.NONE` if the load method of a replay makes no api
    calls.

    :data:`~.RatelimitWeight.LIGHT` if the load method of a replay makes only
    light api calls (anything but ``get_replay``).

    :data:`~.RatelimitWeight.HEAVY` if the load method of a replay makes any
    heavy api calls (``get_replay``).

    Notes
    -----
    This value currently has no effect in circlecore and is reserved for future
    functionality.
    """
    NONE  = "None"
    LIGHT = "Light"
    HEAVY = "Heavy"


class Key(IntFlag):
    M1    = 1 << 0
    M2    = 1 << 1
    K1    = 1 << 2
    K2    = 1 << 3
    SMOKE = 1 << 4


KEY_MASK = int(Key.M1) | int(Key.M2)


def convert_statistic(stat, mods, *, to):
    """
    Converts a game statistic to either its unconverted or converted form,
    depending on ``to``.

    Parameters
    ----------
    stat: float
        The statistic to convert.
    mods: Mod
        The mods the replay was played with. Only ``Mod.DT`` and ``Mod.HT``
        will affect the statistic conversion.
    to: {"cv", "ucv"}
        ``cv`` if the statistic should be converted to its converted form, and
        ``ucv`` if the statistic should be converted to its unconverted form.

    Notes
    -----
    This method is intended for any statistic that is modified from what we
    expect by ``Mod.DT`` or ``Mod.HT`` being applied (ie changing the game
    clock speed). This includes ur (unstable rate) and median frametime
    (time between frames).
    """
    check_param(to, ["cv", "ucv"])

    conversion_factor = 1

    if Mod.DT in mods:
        conversion_factor = (1 / 1.5)
    elif Mod.HT in mods:
        conversion_factor = (1 / 0.75)

    if to == "cv":
        return stat * conversion_factor
    return stat / conversion_factor


def order(replay1, replay2):
    """
    An ordered tuple of the given replays. The first element is the earlier
    replay, and the second element is the later replay.

    Parameters
    ----------
    replay1: Replay
        The first replay to order.
    replay2: Replay
        The second replay to order.

    Returns
    -------
    (Replay, Replay)
        The first element is the earlier replay, and the second element is the
        later replay.
    """
    if not replay1.timestamp or not replay2.timestamp:
        raise ValueError("Both replay1 and replay2 must provide a timestamp. "
            "Replays without a timestamp cannot be ordered.")
    # assume they're passed in order (earliest first); if not, switch them
    order = (replay1, replay2)
    if replay2.timestamp < replay1.timestamp:
        order = tuple(reversed(order))
    return order


def replay_pairs(replays, replays2=None):
    """
    A list of pairs of replays which can be compared against each other to cover
    all cases of replay stealing in ``replays`` and/or ``replays2``.

    If ``replays2`` is not passed (the default), this is a list of 2-tuples
    which are pairs of replays in ``replays``, where each replay will be paired
    with every other replay exactly once.

    If ``replays2`` is passed, this is a list of 2-tuples which are pairs of
    replays in where one replay is from ``replays``, the other is from
    ``replays2``, and every replay in ``replays`` is paired against every replay
    in ``replays2`` (but not against other replays in ``replays``).

    Returns
    -------
    Iterable[(Replay, Replay)]
        The first element is the earlier replay, and the second element is the
        later replay.

    Notes
    -----
    This is equivalent to ``itertools.combinations(replays, 2)`` if ``replays2``
    is ``None`` or the empty list, and ``itertools.product(replays, replays2)``
    otherwise.
    """
    if not replays2:
        return combinations(replays, 2)
    return product(replays, replays2)


def check_param(param, options):
    if param not in options:
        raise ValueError(f"Expected one of {','.join(options)}. Got {param}")


def fuzzy_mods(required_mod, optional_mods):
    """
    All mod combinations where each mod in ``optional_mods`` is allowed to be
    present or absent.

    If you don't want any mods to be required, pass ``Mod.NM`` as your
    ``required_mod``.

    Parameters
    ----------
    required_mod: class:`~circleguard.mod.ModCombination`
        What mod to require be present for all mods.
    optional_mods = [class:`~circleguard.mod.ModCombination`]
        What mods are allowed, but not required, to be present.

    Examples
    --------
    >>> fuzzy_mods(Mod.HD, [Mod.DT])
    [HD, HDDT]
    >>> fuzzy_mods(Mod.HD, [Mod.EZ, Mod.DT])
    [HD, HDDT, HDEZ, HDDTEZ]
    >>> fuzzy_mods(Mod.NM, [Mod.EZ, Mod.DT])
    [NM, DT, EZ, DTEZ]
    """

    all_mods = []
    for mods in powerset(optional_mods):
        final_mod = required_mod
        for mod in mods:
            final_mod = final_mod + mod
        all_mods.append(final_mod)

    return all_mods


def powerset(iterable):
    """
    The powerset of an iterable.

    Examples
    --------
    >>> powerset([1,2,3])
    [(), (1,), (2,), (3,), (1, 2), (1, 3), (2, 3), (1, 2, 3)]

    Notes
    -----
    https://stackoverflow.com/a/1482316
    """
    s = list(iterable)
    return chain.from_iterable(combinations(s, r) for r in range(len(s) + 1))


def hitwindow(OD):
    """
    The number of milliseconds before and after a hitobject's time where that
    hitobject can be hit.
    """
    # stable converts OD (and CS), which are originally a float32, to a
    # double and this causes some hitwindows to be messed up when casted to
    # an int so we replicate this
    return int(150 + 50 * (5 - float(np.float32(OD))) / 5)

def hitwindows(OD):
    hitwindow_50 = hitwindow(OD)
    hitwindow_100 = (280 - 16 * OD) / 2
    hitwindow_300 = (160 - 12 * OD) / 2

    return (hitwindow_50, hitwindow_100, hitwindow_300)

def hitradius(CS):
    """
    The radius, in osu!pixels (?) of where a hitobject can be hit.
    """
    # attempting to match stable hitradius
    return np.float32(64 * ((1.0 - np.float32(0.7) * (float(np.float32(CS)) - 5) / 5)) / 2) * np.float32(1.00041)

def filter_outliers(arr, bias=1.5):
    """
    Returns ``arr` with outliers removed.

    Parameters
    ----------
    arr: list
        List of numbers to filter outliers from.
    bias: int
        Points in ``arr`` which are more than ``IQR * bias`` away from the first
        or third quartile of ``arr`` will be removed.
    """
    q3, q1 = np.percentile(arr, [75 ,25])
    iqr = q3 - q1
    lower_limit = q1 - (bias * iqr)
    upper_limit = q3 + (bias * iqr)
    return [x for x in arr if lower_limit < x < upper_limit]


TRACE = 5

class ColoredFormatter(Formatter):
    """
    A subclass of :class:`logging.Formatter` that uses ANSI escape codes
    to color different parts of the :class:`logging.LogRecord` when printed to
    the console.

    Notes
    -----
    Adapted from https://stackoverflow.com/a/46482050.
    """

    COLOR_PREFIX = '\033['
    COLOR_SUFFIX = '\033[0m'
    COLOR_MAPPING = {
        "TRACE"    : 90, # bright black
        "DEBUG"    : 94, # bright blue
        "INFO"     : 95, # bright magenta
        "WARNING"  : 31, # red
        "ERROR"    : 91, # bright red
        "CRITICAL" : 41, # white on red bg

        "NAME"     : 32, # green
        "MESSAGE"  : 93, # bright yellow
        "FILENAME" : 92, # bright green
        "LINENO"   : 91  # bright red
    }

    def __init__(self, patern):
        Formatter.__init__(self, patern)
        self.colored_log = "{prefix}{{color}}m{{msg}}{suffix}".format(
                    prefix=self.COLOR_PREFIX, suffix=self.COLOR_SUFFIX)

    def format(self, record):
        # c as in colored, not as in copy
        c_record = copy(record)

        # logging's choice of camelCase, not mine
        threadName = c_record.threadName
        color = self.COLOR_MAPPING["NAME"]
        c_threadName = self.colored_log.format(color=color, msg=threadName)

        levelname = c_record.levelname
        color = self.COLOR_MAPPING.get(levelname, 37) # default to white
        c_levelname = self.colored_log.format(color=color, msg=levelname)

        name = c_record.name
        color = self.COLOR_MAPPING["NAME"]
        c_name = self.colored_log.format(color=color, msg=name)

        message = c_record.msg
        color = self.COLOR_MAPPING["MESSAGE"]
        c_msg = self.colored_log.format(color=color, msg=message)

        filename = c_record.filename
        color = self.COLOR_MAPPING["FILENAME"]
        c_filename = self.colored_log.format(color=color, msg=filename)

        lineno = c_record.lineno
        color = self.COLOR_MAPPING["LINENO"]
        c_lineno = self.colored_log.format(color=color, msg=lineno)

        c_record.threadName = c_threadName
        c_record.levelname = c_levelname
        c_record.name = c_name
        c_record.msg = c_msg
        c_record.filename = c_filename
        c_record.lineno = c_lineno

        return Formatter.format(self, c_record)



class GameVersion(int):
    """
    Information about the version of osu! a
    :class:`~circleguard.loadables.Replay` was played on.

    Parameters
    ----------
    version: int
        The version of osu! to represent.

    concrete: bool
        Whether ``version`` is the actual version of osu! the replay was played
        on (in which case ``concrete`` should be ``True``), or just an
        approximation of the version of osu! the replay was played on (in which
        case ``concrete`` should be ``False``).
        |br|
        If the version is an approximation, you are not restricted to actual
        released versions of the game (for instance, osu! has no 20200908
        version of the game, only 20200831), but may use any day of any month
        of any year as your version. Circleguard will see that your version
        is just an estimate (as ``concrete`` will be ``False``), and act
        accordingly.

    Notes
    -----
    osu!'s versioning scheme uses a variation of Calender Versioning
    (https://calver.org/), which means that a release's version is the day that
    release was released. A new version pushed out on 09/08/2020 (MM/DD/YYYY),
    for instance, would have a version of 20200809 (YYYYMMDD).
    |br|
    We provide :func:`~circleguard.game_version.GameVersion.from_datetime` as a
    convenience for when you have a :class:`datetime.datetime` object
    representing the day an osu! version was released, and want to create a
    ``GameVersion`` from that.
    |br|
    This class subclasses ``int`` so consumers don't need to know or expect a
    special class when they access ``replay.game_version``. For instance, to
    get the numeric value of the game version, they would have to do
    ``replay.game_version.version`` as opposed to ``replay.game_version`` here.
    """
    def __new__(cls, version, concrete):
        ret = int.__new__(GameVersion, version)
        ret.concrete = concrete
        return ret

    @staticmethod
    def from_datetime(datetime, concrete):
        """
        Provided as a convenience for converting a :class:`datetime.datetime`
        object to a ``GameVersion`` object.

        Parameters
        ----------
        datetime: :class:`datetime.datetime`
            The datetime to convert to a ``GameVersion`` object.
        concrete: bool
            Whether this version is concrete (ie, fully accurate) or not (ie,
            just an estimate of the replay's actual version).

        Returns
        -------
        :class:`~.GameVersion`
            The result of converting ``datetime`` to a ``GameVersion``.
        """
        version = int(datetime.strftime("%Y%m%d"))
        return GameVersion(version, concrete)

    def available(self):
        """
        Whether we can provide any information about the osu! version a replay
        was played on, whether that is a fully accurate version or just an
        estimate of the version.
        """
        return self != -1


class NoGameVersion(GameVersion):
    """
    Used when a :class:`~circleguard.loadables.Replay` has no information about
    its version, and cannot even estimate its version.
    """
    def __new__(cls):
        return super().__new__(NoGameVersion, -1, None)



import logging
from tempfile import TemporaryDirectory
from typing import Iterable, Union, Tuple
import weakref

from slider import Library, Beatmap

from circleguard.loader import Loader
from circleguard.investigations import Investigations, Snap
from circleguard.judgment import Judgment, Hit
from circleguard.utils import convert_statistic, check_param
from circleguard.loadables import (Map, User, MapUser, ReplayMap, ReplayID,
    ReplayPath, ReplayString, ReplayDir)
from circleguard.mod import Mod


class Circleguard:
    """
    Circleguard is the main entry point for using circlecore. In particular, we
    provide methods to calculate various statistics about
    :class:`~circleguard.loadables.Replay`\s:

    * :meth:`~.similarity` - the similarity between two replays.
    * :meth:`~.ur` - the unstable rate of a replay.
    * :meth:`~.snaps` - any unusual snaps in the cursor movement of a replay.
    * :meth:`~.frametime` - the average frametime of a replay.
    * :meth:`~.frametimes` - a list of the time between frames of a replay.
    * :meth:`~.hits` - the locations in a replay where a hitobject is hit.

    We also provide methods to interact with
    :class:`~circleguard.loadables.Loadable`\s:

    * :meth:`~.load` - loads a :class:`~circleguard.loadables.Loadable`.
    * :meth:`~.load_info` - loads the info of a
      :class:`~circleguard.loadables.ReplayContainer`.
    * :meth:`~.Map` - creates a new :class:`~circleguard.loadables.Map` and
      loads its info.
    * :meth:`~.User` - creates a new :class:`~circleguard.loadables.User` and
      loads its info.
    * :meth:`~.MapUser` - creates a new :class:`~circleguard.loadables.MapUser`
      and loads its info.
    * :meth:`~.ReplayDir` - creates a new
      :class:`~circleguard.loadables.ReplayDir` and loads its info.

    Parameters
    ----------
    key: str
        A valid api key. Can be retrieved from https://osu.ppy.sh/p/api/.
    db_path: str or :class:`os.PathLike`
        The path to the database file to read and write cached replays. If the
        path does not exist, a fresh database will be created there.
        If ``None``, no replays will be cached or loaded from cache.
    slider_dir: str or :class:`os.PathLike`
        The path to the directory used by :class:`slider.library.Library` to
        store beatmaps. If ``None``, a temporary directory will be created for
        :class:`slider.library.Library` and subsequently destroyed when this
        :class:`~Circleguard` object is garbage collected.
    loader: :class:`~circleguard.loader.Loader`
        An instance of :class:`~circleguard.loader.Loader` or a subclass
        thereof, which will be used instead of creating a loader if passed.
    """
    DEFAULT_ANGLE = 10
    DEFAULT_DISTANCE = 8
    # A good balance between speed and accuracy when calculating correlation.
    DEFAULT_CHUNKS = 5

    # We provide suggested "cheated thresholds" for comparisons, as those
    # algorithms are our own invention and consumers should not be expected to
    # know how they work at anything but a high level.
    # We do not, however, provide thresholds for any other statistic, as they
    # are usually gameplay statistics that consumers should understand well
    # enough to define their own threshold.
    SIM_LIMIT = 17
    CORR_LIMIT = 0.99

    def __init__(self, key, db_path=None, slider_dir=None, loader=None,
        cache=True
    ):
        self.log = logging.getLogger(__name__)

        # allow for people to pass their own loader implementation/subclass.
        # Mostly exposed for circleguard (the gui).
        if loader:
            self.loader = loader
        else:
            self.loader = Loader(key, db_path, write_to_cache=cache)

        if slider_dir:
            self.library = Library(slider_dir)
        else:
            # If slider dir wasn't passed, use a temporary library which will
            # effectively cache beatmaps for just this cg instance.
            # Have to keep a reference to this dir or the folder gets deleted.
            self.slider_dir = TemporaryDirectory()
            self.library = Library(self.slider_dir.name)
            # clean up our library (which resides in a temporary dir) or else
            # garbage collection of this cg object (and subsequently the
            # temp dir and library) will cause an error to be thrown. This
            # happens because the temp dir's finalizer is called first, which
            # tries to remove the directory, but it can't because the library's
            # sql connection to the db file in that dir is still alive, and a
            # PermissionError is thrown. We need to close the library before
            # the temp dir is finalized.
            # Errors that happen during garbage collection are ignored I
            # believe, so this only fixes the error message appearing (which is
            # still a good thing to do) rather than actually fixing any programs
            # that broke because of this.
            self._finalizer = weakref.finalize(self, self._cleanup,
                self.library)


    def similarity(self, replay1, replay2, method="similarity",
        num_chunks=DEFAULT_CHUNKS, mods_unknown="best") -> \
        Union[float, Tuple[float]]:
        """
        The similarity between ``replay1`` and ``replay2``.

        Parameters
        ----------
        replay1: :class:`~circleguard.loadables.Replay`
            The replay to compare against ``replay2``.
        replay2: :class:`~circleguard.loadables.Replay`
            The replay to compare against ``replay1``.
        method: {"similarity", "correlation"}
            What method to use to calculate the similarity between the replays.
            |br|
            ``similarity`` is (roughly speaking) the average distance
            between the two replays in pixels. A replay compared to itself (or
            an exact copy) has a similarity of 0. See
            :data:`~circleguard.Circleguard.SIM_LIMIT` for a suggested number
            where similarities below this number indicate a stolen replay.
            |br|
            ``correlation`` is a signal-processing metric which measures how
            similar two signals (or replays, in our case) are. Correlation also
            takes into account time shifts, so a replay which is a perfect copy
            of another replay but consistently lags 10 ms behind will still have
            a perfect correltion of ``1``. The higher correlation, the more
            correlated (or similar) the two replays are. See
            :data:`~circleguard.Circleguard.CORR_LIMIT` for a suggested number
            where correlations above this number indicate a stolen replay.
        num_chunks: int
            How many chunks to split the replay into when comparing. This
            parameter only has an affect if ``method`` is ``correlation``.
            Note that runtime increases linearly with the number of chunks.
        mods_unknown: {"best", "both"}
            What to do if one or both of ``replay1`` and ``replay2`` do not
            know what mods they were played with. In this case, the similarity
            will be computed twice, both with no modifications and with
            ``Mod.HR`` applied to ``replay1``.
            |br|
            If ``best`` is passed, the best (that is, lowest if ``method`` is
            ``similarity`` and highest if ``method`` is ``correlation``)
            similarity of these two calculations is returned.
            |br|
            If ``both`` is passed, a tuple with two floats is returned. The
            first element is the similarity with no modifications, and the
            second element is the similarity with ``Mod.HR`` applied to
            ``replay1``.

        Returns
        -------
        float
            If ``method`` is ``similarity``, this is the similarity of the two
            replays. If ``method`` is ``correlation``, this is the correlation
            between the two replays.
        (float, float)
            If ``mods_unknown`` is ``both``, a tuple with two floats
            is returned. The first element is the similarity with no
            modifications, and the second element is the similarity with
            ``Mod.HR`` applied to ``replay1``. See the documentation for the
            ``mods_unknown`` parameter for more information.
        """
        check_param(method, ["similarity", "correlation"])
        check_param(mods_unknown, ["best", "both"])

        self.load(replay1)
        self.load(replay2)
        return Investigations.similarity(replay1, replay2, method, num_chunks,
            mods_unknown)


    def ur(self, replay, cv=True, beatmap=None, adjusted=False) -> float:
        """
        The unstable rate of ``replay``.

        Parameters
        ----------
        replay: :class:`~circleguard.loadables.Replay`
            The replay to calculate the ur of.
        cv: bool
            Whether to return the converted or unconverted ur. The converted ur
            is returned by default.
        beatmap: :class:`slider.beatmap.Beatmap`
            The beatmap to use to calculate ur for the ``replay``, instead of
            retrieving a beatmap from the replay itself.
            |br|
            This parameter is provided primarily as an optimization for when you
            already have the replay's beatmap, to avoid re-retrieving it in this
            method.
        adjusted: boolean
            Whether to calculate "adjusted" ur. Adjusted ur filters outlier hits
            before calculating ur and can result in a more accurate ur for some
            replays.

        Returns
        -------
        float
            The ur of the replay.
        """
        self.load(replay)

        beatmap = beatmap or self.beatmap(replay)
        if not beatmap:
            raise ValueError("The ur of a replay that does not know what map "
                "it was set on cannot be calculated")

        ur = Investigations.ur(replay, beatmap, adjusted)
        if cv:
            ur = convert_statistic(ur, replay.mods, to="cv")

        return ur


    def snaps(self, replay, max_angle=DEFAULT_ANGLE,
        min_distance=DEFAULT_DISTANCE, only_on_hitobjs=True,
        beatmap=None) -> Iterable[Snap]:
        """
        Finds any snaps (sudden, jerky movement) in ``replay``.

        Parameters
        ----------
        replay: :class:`~circleguard.loadables.Replay`
            The replay to find snaps in.
        max_angle: float
            Consider only (a,b,c) where ``∠abc < max_angle``
        min_distance: float
            Consider only (a,b,c) where ``|ab| > min_distance`` and
            ``|bc| > min_distance``.
        only_on_hitobjs: bool
            Whether to only return snaps that occur on a hitobject.
        beatmap: :class:`slider.beatmap.Beatmap`
            The beatmap to use to calculate snaps for the ``replay``, instead of
            retrieving a beatmap from the replay itself. This is only used when
            ``only_on_hitobjs`` is true, since the beatmap is not necessary
            otherwise.
            |br|
            This parameter is provided primarily as an optimization for when you
            already have the replay's beatmap, to avoid re-retrieving it in this
            method.

        Returns
        -------
        list[Snap]
            The snaps of the replay.

        Notes
        -----
        Specifically, this function calculates the angle between each set of
        three points (a,b,c) and finds points where this angle is extremely
        acute and neither ``|ab|`` or ``|bc|`` are small.

        By default, only snaps which occur on a hitobject are returned. This is
        to reduce false positives from spinners, driver issues, or lifting the
        pen off the tablet and back on again.
        """
        self.load(replay)

        beatmap_ = None
        if only_on_hitobjs:
            beatmap_ = beatmap or self.beatmap(replay)
            if not beatmap_:
                raise ValueError("The snaps of a replay that does not know "
                    "what map it was set on cannot be filtered to include only "
                    "snaps on hit objects. If you cannot retrieve the beatmap, "
                    "you can pass ``only_on_hitobjs=False`` to avoid requiring "
                    "a beatmap.")

        return Investigations.snaps(replay, max_angle, min_distance, beatmap_)


    def frametime(self, replay, cv=True, mods_unknown="raise") -> float:
        """
        The median frametime (in ms) of ``replay``.

        Parameters
        ----------
        replay: :class:`~circleguard.loadables.Replay`
            The replay to calculate the median frametime of.
        cv: bool
            Whether to return the converted or unconverted frametime. The
            converted frametime is returned by default.
        mods_unknown: {"raise", "dt", "nm", "ht"}
            What to do if ``replay`` does not know what mods it was played with,
            and ``cv`` is ``True``.
            |br|
            If ``raise``, a ValueError will be raised.
            |br|
            If ``dt``, the frametime will be converted as if the replay was
            played with ``Mod.DT``.
            |br|
            If ``nm``, the frametime will be converted as if the replay was
            played  with ``Mod.NM`` (that is, not converted at all).
            |br|
            If ``ht``, the frametime will be converted as if the replay was
            played with ``Mod.HT``.

        Returns
        -------
        float
            The median frametime (in ms) of the replay.
        """
        check_param(mods_unknown, ["raise", "dt", "nm", "ht"])

        self.load(replay)
        frametime = Investigations.frametime(replay)
        if cv:
            if replay.mods:
                mods = replay.mods
            elif mods_unknown == "dt":
                mods = Mod.DT
            elif mods_unknown == "nm":
                mods = Mod.NM
            elif mods_unknown == "ht":
                mods = Mod.HT
            else:
                raise ValueError("The frametime of a replay that does not know "
                    "with what mods it was set with cannot be converted. Pass "
                    "a different option to frametime(..., mods_unknown=) if "
                    "you would like to provide a default mod for conversion.")
            frametime = convert_statistic(frametime, mods, to="cv")

        return frametime

    def frametimes(self, replay, cv=True, mods_unknown="raise") \
        -> Iterable[float]:
        """
        The time (in ms) between each frame in ``replay``.

        Parameters
        ----------
        replay: :class:`~circleguard.loadables.Replay`
            The replay to calculate the time between each frame of.
        cv: bool
            Whether to return the converted or unconverted frametimes. The
            converted frametimes is returned by default.
        mods_unknown: {"raise", "dt", "nm", "ht"}
            What to do if ``replay`` does not know what mods it was played with,
            and ``cv`` is ``True``.
            |br|
            If ``raise``, a ValueError will be raised.
            |br|
            If ``dt``, the frametime swill be converted as if the replay was
            played with ``Mod.DT``.
            |br|
            If ``nm``, the frametimes will be converted as if the replay was
            played  with ``Mod.NM`` (that is, not converted at all).
            |br|
            If ``ht``, the frametimes will be converted as if the replay was
            played with ``Mod.HT``.

        Returns
        -------
        [float]
            The time (in ms) between each frame of the replay.
            |br|
            The first element of this array corresponds to the time between the
            first and second frame, the second element to the time between the
            second and third frame, etc.
        """
        self.load(replay)
        frametimes = Investigations.frametimes(replay)
        if cv:
            if replay.mods:
                mods = replay.mods
            elif mods_unknown == "dt":
                mods = Mod.DT
            elif mods_unknown == "nm":
                mods = Mod.NM
            elif mods_unknown == "ht":
                mods = Mod.HT
            else:
                raise ValueError("The frametimes of a replay that does not "
                    "know with what mods it was set with cannot be converted. "
                    "Pass one of ``{\"dt\", \"nm\", \"ht\"}`` for "
                    "``mods_unknown``if you would like to provide a default "
                    "mod for conversion.")
            frametimes = convert_statistic(frametimes, mods, to="cv")
        return frametimes

    def hits(self, replay, within=None, beatmap=None) -> Iterable[Hit]:
        """
        The locations in the replay where a hitobject is hit.

        Parameters
        ----------
        replay: :class:`~circleguard.loadables.Replay`
            The replay to calculate the hits of.
        within: float
            If passed, only the hits which are ``within`` pixels or less away
            from the edge of the hitobject which they hit will be returned.
            Otherwise, all hits are returned.
        beatmap: :class:`slider.beatmap.Beatmap`
            The beatmap to use to calculate hits for the ``replay``, instead of
            retrieving a beatmap from the replay itself.
            |br|
            This parameter is provided primarily as an optimization for when you
            already have the replay's beatmap, to avoid re-retrieving it in this
            method.

        Returns
        -------
        list[:class:`~circleguard.judgment.Judgment`]
            The hits of the replay.

        Notes
        -----
        In osu!lazer terminology, hits are equivalent to judgements, but
        without misses.
        """
        self.load(replay)

        beatmap = beatmap or self.beatmap(replay)
        if not beatmap:
            raise ValueError("The hits of a replay that does not know what map "
                "it was set on cannot be calculated.")

        hits = Investigations.hits(replay, beatmap)

        if not within:
            return hits

        hits = [hit for hit in hits if hit.within(within)]
        return hits

    def judgments(self, replay, beatmap=None) -> Iterable[Judgment]:
        """
        The locations in the replay where a hitobject is hit or missed.
        Judgments are marked as either misses, 50s, 100s, or 300s.

        Parameters
        ----------
        replay: :class:`~circleguard.loadables.Replay`
            The replay to calculate the judgments of.
        beatmap: :class:`slider.beatmap.Beatmap`
            The beatmap to use to calculate judgments for the ``replay``,
            instead of retrieving a beatmap from the replay itself.
            |br|
            This parameter is provided primarily as an optimization for when you
            already have the replay's beatmap, to avoid re-retrieving it in this
            method.

        Returns
        -------
        list[:class:`~circleguard.judgment.Judgment`]
            The judgments of the replay.
        """
        self.load(replay)

        beatmap = beatmap or self.beatmap(replay)
        if not beatmap:
            raise ValueError("The judgments of a replay that does not know "
                "what map it was set on cannot be calculated.")

        return Investigations.judgments(replay, beatmap)

    def frametime_graph(self, replay, cv=True, figure=None,
        show_expected_frametime=True):
        """
        Uses matplotlib to create a graph of the frametimes of the replay.

        Parameters
        ----------
        replay: :class:`~circleguard.loadables.Replay`
            The replay to graph the frametimes of.
        cv: bool
            Whether the frametimes should be converted before being graphed.
        figure: :class:`matplotlib.figure.Figure`
            If passed, this figure will be used instead of creating a new one
            with pyplot. Using this parameter is not recommended for normal
            usage. It is exposed to allow circleguard (the gui) to use this
            method, as matplotlib's pyqt compatability layer adds some
            complications which this works around.
        show_expected_frametime: bool
            Whether to show a vertical line where we expect the average
            frametime to be.

        Returns
        -------
        :module:`matplotlib.pyplot` or :class:`matplotlib.figure.Figure`
            Matplotlib's pyplot module for ease of use, so you can call
            :meth:`matplotlib.pyplot.show` on the return value of this function
            to display the graph.
            |br|
            If ``figure`` is passed, the return value is instead the passed
            figure after being modified by the frametime graph.
        """
        # we raise an ImportError if the consumer doesn't have matplotlib
        # installed, which is why we have to import it only when this function
        # is called.
        from circleguard.frametime_graph import FrametimeGraph
        from matplotlib import pyplot
        self.load(replay)
        frametime_graph = FrametimeGraph(replay, cv, figure,
            show_expected_frametime)
        # strictly speaking, I don't think this return is necessary -
        # ``FrametimeGraph`` modifies the passed figure on instantiation,
        # so consumers could just use the figure they passed in instead of the
        # return value here. Returning it anyway feels better though.
        return frametime_graph.figure if figure else pyplot


    def load(self, loadable):
        """
        Loads the ``loadable``.

        Parameters
        ----------
        loadable: :class:`~circleguard.loadables.Loadable`
            The loadable to load.

        Notes
        -----
        This is identical to calling ``loadable.load(cg.loader, cg.cache)``.
        """
        loadable.load(self.loader, self.cache)


    def load_info(self, replay_container):
        """
        Loads the info of the ``replay_container``.

        Parameters
        ----------
        replay_container: :class:`~circleguard.loadables.ReplayContainer`
            The replay container to load.

        Notes
        -----
        This is identical to calling
        ``replay_container.load_info(cg.loader)``.
        """
        replay_container.load_info(self.loader)

    def beatmap_available(self, replay):
        return replay.beatmap_available(self.library)

    # TODO remove in core 6.0.0
    map_available = beatmap_available

    def Map(self, beatmap_id, span, mods=None, cache=None, load=False) -> Map:
        """
        Returns a new, info-loaded :class:`~circleguard.loadables.Map`.

        Notes
        -----
        This function is provided as a convenience for when you want to create a
        ``Map`` and load its info immediately. A common occurrence in using
        ``Circleguard`` is to want to instantiate a ``Map`` and immediately
        iterate over it to operate on its replays. However, this ``Map`` must
        be info loaded before it can be iterated over, so this function does
        that info loading for you.

        >>> # good
        >>> m = cg.Map(221777, "1-2")
        >>> for replay in m:
        >>>     ...

        >>> # bad
        >>> m = Map(221777, "1-2")
        >>> cg.load_info(m)
        >>> for replay in m:
        >>>     ...
        """
        m = Map(beatmap_id, span, mods, cache)
        self.load(m) if load else self.load_info(m)
        return m

    def User(self, user_id, span, mods=None, cache=None,
        available_only=True, load=False) -> User:
        """
        Returns a new, info-loaded :class:`~circleguard.loadables.User`.

        Notes
        -----
        This function is provided as a convenience for when you want to create a
        ``User`` and load its info immediately. A common occurrence in using
        ``Circleguard`` is to want to instantiate a ``User`` and immediately
        iterate over it to operate on its replays. However, this ``User`` must
        be info loaded before it can be iterated over, so this function does
        that info loading for you.

        >>> # good
        >>> u = cg.User(124493, "1-2")
        >>> for replay in u:
        >>>     ...

        >>> # bad
        >>> u = User(124493, "1-2")
        >>> cg.load_info(u)
        >>> for replay in u:
        >>>     ...
        """
        u = User(user_id, span, mods, cache, available_only)
        self.load(u) if load else self.load_info(u)
        return u

    def MapUser(self, beatmap_id, user_id, span=Loader.MAX_MAP_SPAN, cache=None,
        available_only=True, load=False) -> MapUser:
        """
        Returns a new, info-loaded :class:`~circleguard.loadables.MapUser`.

        Notes
        -----
        This function is provided as a convenience for when you want to create a
        ``MapUser`` and load its info immediately. A common occurrence in using
        ``Circleguard`` is to want to instantiate a ``MapUser`` and immediately
        iterate over it to operate on its replays. However, this ``MapUser``
        must be info loaded before it can be iterated over, so this function
        does that info loading for you.

        >>> # good
        >>> mu = cg.MapUser(124493, 129891)
        >>> for replay in mu:
        >>>     ...

        >>> # bad
        >>> mu = MapUser(124493, 129891)
        >>> cg.load_info(mu)
        >>> for replay in mu:
        >>>     ...
        """
        mu = MapUser(beatmap_id, user_id, span, cache, available_only)
        self.load(mu) if load else self.load_info(mu)
        return mu

    def ReplayDir(self, path, cache=None, load=False) -> ReplayDir:
        """
        Returns a new, info-loaded :class:`~circleguard.loadables.ReplayDir`.

        Notes
        -----
        This function is provided as a convenience for when you want to create a
        ``ReplayDir`` and load its info immediately. A common occurrence in using
        ``Circleguard`` is to want to instantiate a ``ReplayDir`` and immediately
        iterate over it to operate on its replays. However, this ``ReplayDir``
        must be info loaded before it can be iterated over, so this function
        does that info loading for you.

        >>> # bad
        >>> r_dir = cg.ReplayDir("/Users/tybug/Desktop/replays")
        >>> for replay in r_dir:
        >>>     ...

        >>> # good
        >>> r_dir = ReplayDir("/Users/tybug/Desktop/replays")
        >>> cg.load_info(r_dir)
        >>> for replay in r_dir:
        >>>     ...
        """
        r_dir = ReplayDir(path, cache)
        self.load(r_dir) if load else self.load_info(r_dir)
        return r_dir


    def ReplayMap(self, beatmap_id, user_id, mods=None, cache=None, info=None) \
        -> ReplayMap:
        """
        Returns a new, loaded :class:`~circleguard.loadables.ReplayMap`.

        Notes
        -----
        This function is provided as a convenience for when you want to create a
        ``ReplayMap`` and load it immediately. Loading can be an expensive
        operation which is why this does not occur by default.
        """
        r = ReplayMap(beatmap_id, user_id, mods, cache, info)
        self.load(r)
        return r

    def ReplayPath(self, path, cache=None) -> ReplayPath:
        """
        Returns a new, loaded :class:`~circleguard.loadables.ReplayPath`.

        Notes
        -----
        This function is provided as a convenience for when you want to create a
        ``ReplayPath`` and load it immediately. Loading can be an expensive
        operation which is why this does not occur by default.
        """
        r = ReplayPath(path, cache)
        self.load(r)
        return r

    def ReplayString(self, replay_data_str, cache=None) -> ReplayString:
        """
        Returns a new, loaded :class:`~circleguard.loadables.ReplayString`.

        Notes
        -----
        This function is provided as a convenience for when you want to create a
        ``ReplayString`` and load it immediately. Loading can be an expensive
        operation which is why this does not occur by default.
        """
        r = ReplayString(replay_data_str, cache)
        self.load(r)
        return r

    def ReplayID(self, replay_id, cache=None) -> ReplayID:
        """
        Returns a new, loaded :class:`~circleguard.loadables.ReplayID`.

        Notes
        -----
        This function is provided as a convenience for when you want to create a
        ``ReplayID`` and load it immediately. Loading can be an expensive
        operation which is why this does not occur by default.
        """
        r = ReplayID(replay_id, cache)
        self.load(r)
        return r

    def beatmap(self, replay) -> Beatmap:
        """
        The beatmap the replay was played on, or ``None`` if the replay doesn't
        know what beatmap it was played on.

        Returns
        -------
        :class:`slider.beatmap.Beatmap`
            The beatmap this replay was played on.
        None
            If the replay doesn't know what beatmap it was played on.
        """
        self.load(replay)
        return replay.beatmap(self.library)

    @property
    def cache(self):
        return self.loader.write_to_cache

    @cache.setter
    def cache(self, cache):
        self.loader.write_to_cache = cache

    @classmethod
    def _cleanup(cls, library):
        # see the call to this method for documentation on why this method
        # exists.
        library.close()

class KeylessCircleguard(Circleguard):
    """
    A :class:`~.Circleguard` for when you do not have access to an api key, but
    have loaded :class:`~circleguard.loadables.Loadable`\s that you want to
    perform operations on. It should go without saying that instances of this
    class cannot do anything that requires api access.

    ``KeylessCircleguard``s may also load ``ReplayPath``\s and
    ``ReplayString``\s, but some attributes of these replays will not be able to
    be accessed, as they require api access (such as user id or map id).

    Parameters
    ----------
    db_path: str or :class:`os.PathLike`
        The path to the database file to read and write cached replays. If the
        path does not exist, a fresh database will be created there.
        If `None`, no replays will be cached or loaded from cache.
    slider_dir: str or :class:`os.PathLike`
        The path to the directory used by :class:`slider.library.Library` to
        store beatmaps. If `None`, a temporary directory will be created for
        :class:`slider.library.Library` and subsequently destroyed when this
        :class:`~Circleguard` object is garbage collected.
    loader: :class:`~circleguard.loader.Loader`
        An instance of :class:`~circleguard.loader.Loader` or a subclass
        thereof, which will be used instead of creating a loader if passed.
    """

    def __init__(self, db_path=None, slider_dir=None, cache=True):
        super().__init__("INVALID_KEY", db_path, slider_dir, cache=cache)

    def load(self, loadable):
        loadable.load(None, self.cache)

    def load_info(self, replay_container):
        replay_container.load_info(None)


def set_options(*, loglevel=None):
    """
    Set global options for circlecore.

    Parameters
    ---------
    logevel: int
        What level to log at. Circlecore follows standard python logging
        levels, with an added level of TRACE with a value of 5 (lower than
        debug, which is 10). The value passed to loglevel is passed directly to
        the setLevel function of the circleguard root logger. WARNING by
        default. For more information on log levels, see the standard python
        logging lib.
    """
    if loglevel is not None:
        logging.getLogger("circleguard").setLevel(loglevel)



from tests.utils import CGTestCase

from circleguard import Mod, fuzzy_mods

class TestLoader(CGTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()

    def test_mod_string_parsing(self):
        # one normal, one "special" (nc and pf), and one multimod mod
        self.assertEqual(Mod("HD"), Mod.HD)
        self.assertEqual(Mod("NC"), Mod.NC)
        self.assertEqual(Mod("SOHDDT"), Mod.HD + Mod.DT + Mod.SO)

        self.assertRaises(ValueError, lambda: Mod("DTH"))
        self.assertRaises(ValueError, lambda: Mod("DH"))

    def test_mod_str_list_parsing(self):
        self.assertEqual(Mod(["HD"]), Mod.HD)
        self.assertEqual(Mod(["NC"]), Mod.NC)
        self.assertEqual(Mod(["SO", "HD", "DT"]), Mod.HD + Mod.DT + Mod.SO)
        self.assertEqual(Mod(["SOHD", "DT"]), Mod.HD + Mod.DT + Mod.SO)
        self.assertEqual(Mod(["SOHDDT"]), Mod.HD + Mod.DT + Mod.SO)
        self.assertEqual(Mod(["HD", "SODT"]), Mod.HD + Mod.DT + Mod.SO)

        self.assertRaises(ValueError, lambda: Mod(["DTH"]))
        self.assertRaises(ValueError, lambda: Mod(["DH"]))
        self.assertRaises(ValueError, lambda: Mod(["DH", int()]))

    def test_equality_reflexivity(self):
        self.assertEqual(Mod("NC"), Mod("NC"))

    def test_mod_ordering(self):
        self.assertEqual(Mod("DTHDSO"), Mod("SOHDDT"), "Identical mods ordered differently were not equal")
        self.assertEqual(Mod("DTHR").long_name(), Mod("HRDT").long_name(), "Long name of identical mods ordered differently were not equal")
        self.assertEqual(Mod("SOAPFLEZ").short_name(), Mod("EZSOFLAP").short_name(), "Short name of identical mods ordered differently were not equal")

        self.assertEqual(Mod("HD").short_name(), "HD")
        self.assertEqual(Mod("HR").long_name(), "HardRock")
        self.assertEqual(Mod("DTHR").long_name(), "DoubleTime HardRock")
        self.assertEqual(Mod("HRDT").long_name(), "DoubleTime HardRock")

    def test_fuzzy_mod(self):
        mods = fuzzy_mods(Mod.HD, [Mod.DT, Mod.EZ])
        self.assertListEqual(mods,
            [Mod.HD, Mod.HDDT, Mod.HD + Mod.EZ, Mod.HD + Mod.EZ + Mod.DT])

        mods = fuzzy_mods(Mod.HD, [Mod.DT])
        self.assertListEqual(mods, [Mod.HD, Mod.HD + Mod.DT])

        mods = fuzzy_mods(Mod.NM, [Mod.DT, Mod.EZ])
        self.assertListEqual(mods, [Mod.NM, Mod.DT, Mod.EZ, Mod.DT + Mod.EZ])



# TODO: remove in core 6.0.0, in favor of ``Replay#map_available`` (and
# possibly other mechanisms).
class MapInfo():
    """
    Represents the information necessary to load a beatmap.

    Notes
    -----
    If multiple ways to load a beatmap are known, all ways should be provided
    so consumers can choose the order of ways to load the beatmap.

    If a way to load a beatmap is *not* available, it should be left as
    ``None``.
    """

    def __init__(self, *, map_id=None, path=None):
        self.map_id = map_id
        self.path = path

    def available(self):
        """
        Whether this beatmap can be loaded with the information we have or not.
        """
        return bool(self.map_id) or bool(self.path)



from circleguard import (ReplayPath, ReplayMap, Map, KeylessCircleguard, Mod,
    User, MapUser)

from tests.utils import CGTestCase, RES


class TestReplays(CGTestCase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.kcg = KeylessCircleguard()

    def test_loading_replaypath(self):
        r = ReplayPath(RES / "example_replay.osr")
        self.kcg.load(r)
        self.assertRaises(ValueError, lambda: r.map_id)
        self.assertRaises(ValueError, lambda: r.user_id)
        self.assertEqual(r.mods, Mod.HD + Mod.DT, "Mods was not correct")
        self.assertEqual(r.replay_id, 2029801532, "Replay id was not correct")
        self.assertEqual(r.username, "MarthXT", "Username was not correct")
        self.assertTrue(r.loaded, "Loaded status was not correct")

    def test_loading_other_loadables(self):
        r = ReplayMap(221777, 2757689)
        m = Map(221777, "1")
        u = User(12092800, "1")
        mu = MapUser(221777, 12092800, "1")
        self.assertRaises(ValueError, lambda: self.kcg.load(r))
        self.assertRaises(ValueError, lambda: self.kcg.load(m))
        self.assertRaises(ValueError, lambda: self.kcg.load_info(m))
        self.assertRaises(ValueError, lambda: self.kcg.load(u))
        self.assertRaises(ValueError, lambda: self.kcg.load_info(u))
        self.assertRaises(ValueError, lambda: self.kcg.load(mu))
        self.assertRaises(ValueError, lambda: self.kcg.load_info(mu))



from circleguard import Loader, InvalidKeyException
from tests.utils import CGTestCase, KEY

class TestLoader(CGTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.loader = Loader(KEY)

    def test_loading_map_id(self):
        result = self.loader.map_id("E")
        self.assertEqual(result, 0)

        result = self.loader.map_id("9d0a8fec2fe3f778334df6bdc60b113c")
        self.assertEqual(result, 221777)

    def test_loading_user_id(self):
        result = self.loader.user_id("E")
        self.assertEqual(result, 0)

        result = self.loader.user_id("] [")
        self.assertEqual(result, 13506780)

        result = self.loader.user_id("727")
        self.assertEqual(result, 10750899)

    def test_loading_username(self):
        result = self.loader.username(0)
        self.assertEqual(result, "")

        result = self.loader.username(13506780)
        self.assertEqual(result, "] [")

    def test_incorrect_key(self):
        loader = Loader("incorrect key")
        self.assertRaises(InvalidKeyException, loader.username, 13506780)
        self.assertRaises(InvalidKeyException, loader.user_id, "] [")
        self.assertRaises(InvalidKeyException, loader.map_id, "9d0a8fec2fe3f778334df6bdc60b113c")



# Benchmarks for circleguard. These are not strictly "tests" and will not be run
# by unittest, but it felt appropriate to place them in the tests folder.

import cProfile
from pstats import Stats

from circleguard import Circleguard, ReplayPath
from utils import KEY, RES

cg = Circleguard(KEY)

replay1 = ReplayPath(RES / "legit" / "legit-1.osr")
cg.load(replay1)

def benchmark_ur(with_cache):
    """
    Parameters
    ----------
    with_cache: {"none", "replay", "beatmap", "both"}
        What caches to use when calculating the ur of the replay. If "replay",
        a replay that has already been loaded is used. If "beatmap", the beatmap
        is downloaded on the first call, and is cached thereafter. If "both",
        both of the above caches apply.
    """
    profiler = cProfile.Profile()

    if with_cache == "none":
        num_calls = 10
        def benchmark_func():
            # creating a new cg object is a hacky way to force a new slider
            # cache
            cg = Circleguard(KEY)
            replay = ReplayPath(RES / "legit" / "legit-1.osr")
            cg.ur(replay)
    elif with_cache == "replay":
        num_calls = 10
        def benchmark_func():
            cg = Circleguard(KEY)
            cg.ur(replay1)

    elif with_cache == "beatmap":
        num_calls = 100
        def benchmark_func():
            replay = ReplayPath(RES / "legit" / "legit-1.osr")
            cg.ur(replay)

    elif with_cache == "both":
        num_calls = 100
        def benchmark_func():
            cg.ur(replay1)


    profiler.enable()
    for _ in range(num_calls):
        benchmark_func()

    profiler.disable()

    stats = Stats(profiler)
    print(f"Average time per call with cache \"{with_cache}\": {stats.total_tt / num_calls:3f} ({num_calls} calls)")

if __name__ == "__main__":
    benchmark_ur(with_cache="none")
    benchmark_ur(with_cache="beatmap")
    benchmark_ur(with_cache="replay")
    benchmark_ur(with_cache="both")



from distutils.core import setup
from setuptools import find_packages
import re

with open("README.md", "r") as readme:
    long_description = readme.read()

# https://stackoverflow.com/a/7071358
VERSION = "Unknown"
VERSION_RE = r"^__version__ = ['\"]([^'\"]*)['\"]"

with open("circleguard/version.py") as f:
    match = re.search(VERSION_RE, f.read())
    if match:
        VERSION = match.group(1)
    else:
        raise RuntimeError("Unable to find version string in circleguard/version.py")

setup(
    name="circleguard",
    version=VERSION,
    description="A utilities library for osu!. Provides support for parsing "
        "replays from a file or from the api, as well as support for unstable "
        "rate, hits, similarity, and frametime calculations.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    keywords = ["osu!, python, cheat-detection, replay-stealing, remodding"],
    author="Liam DeVoe",
    author_email="orionldevoe@gmail.com",
    url="https://github.com/circleguard/circlecore",
    download_url = "https://github.com/circleguard/circlecore/tarball/v" + VERSION,
    packages=find_packages(),
    install_requires=[
        "osrparse~=6.0",
        "ossapi~=2.1",
        "wtc==1.2.1",
        "numpy",
        "requests",
        "slider>=0.5.1",
        "scipy"
    ],
    extras_require={
        "graphing": [
            "matplotlib"
        ]
    }
)
